from datetime import datetime, timedelta
import json
from time import mktime
from typing import Dict, List, Tuple
from wsgiref.handlers import format_date_time
from wsgiref.types import StartResponse


class Downlink:
  def __init__(self, start_response: StartResponse):
    self.__start_response = start_response
    self.__body = ''
    self.__status = '200 OK'
    self.__headers: List[Tuple[str, str]] = []
    self.__encoding = 'utf-8'

  def __iter__(self):
    for b in self.__body:
      yield b.encode()

  def header(self, name: str, value: str):
    self.__headers = [h for h in self.__headers if name != h[0]]
    self.__headers.append((name, value))

    return self

  def status(self, code: int):
    self.__status = f'{code} '
    return self

  def text(self, message: str):
    self.header('Content-Type', 'text/plain')
    self.__start_response(self.__status, self.__headers)
    self.__body = message

    return self

  def json(self, data: Dict):
    self.header('Content-Type', 'application/json')
    self.__start_response(self.__status, self.__headers)
    self.__body = json.dumps(data)

    return self
  
  def html(self, data: Dict):
    self.header('Content-Type', 'text/html')
    self.__start_response(self.__status, self.__headers)
    self.__body = json.dumps(data)

    return self
  
  def redirect(self, url: str):
    self.header('Location', url)
    self.status(301)

    return self

  def remember(self, name: str, value: str, expires=datetime.now() + timedelta(days=42)):
    expires = mktime(expires.timetuple())
    expires = format_date_time(expires)
    self.header('Set-Cookie', f'{name}={value}; Expires={expires}')

    return self

  def forget(self, name: str):
    self.remember(name, '', datetime(1970, 1, 1))
    return self



import json
from typing import Dict, List, Literal, Union
from urllib.parse import parse_qs
from wsgiref.types import WSGIEnvironment


class Uplink:
  def __init__(self, environ: WSGIEnvironment):
    self.__environ = environ

  @property
  def method(self) -> Literal['GET', 'POST', 'PUT', 'PATCH', 'DELETE']:
    return self.__environ['REQUEST_METHOD']

  @property
  def query(self) -> Dict[str, List[str]]:
    return parse_qs(self.__environ['QUERY_STRING'])

  @property
  def body(self) -> Union[str, Dict, List[Dict]]:
    c_length = int(self.__environ['CONTENT_LENGTH'])
    c_type = self.__environ['CONTENT_TYPE']
    c_data = self.__environ['wsgi.input'].read(c_length).decode()

    match c_type:
      case 'application/json':
        return json.loads(c_data)
      case _:
        return c_data



from typing import Callable, List
from wsgiref.simple_server import make_server
from wsgiref.types import StartResponse, WSGIEnvironment
from .datalink import Datalink
from .uplink import Uplink
from .downlink import Downlink


class Satellite:
  def __init__(self, port = 3000, hostname = 'localhost'):
    self.__server = make_server(hostname, port, lambda e, s: self.__router(e, s))
    self.__datalinks: List[Datalink] = []
    self.__hostname = hostname
    self.__port = port
  
  def transmit(self):
    try:
      print(f'Transmitting on http://{self.__hostname}:{self.__port}')
      self.__server.serve_forever()
    except KeyboardInterrupt:
      print('\nTransmission terminated')

  def __router(self, environ: WSGIEnvironment, start_response: StartResponse):
    downlink = Downlink(start_response)

    for datalink in self.__datalinks:
      if datalink.match(environ['PATH_INFO']):
        uplink = Uplink(environ)
        return datalink.handler(uplink, downlink)
      
    return downlink.status(404).json({'error': 'Houston, we have a problem!'})
  
  def datalink(self, route: str):
    def inner(handler: Callable[[Uplink, Downlink], Downlink]):
      self.__datalinks.append(Datalink(route, handler))

    return inner



from dataclasses import dataclass
from typing import Callable
from .uplink import Uplink
from .downlink import Downlink


@dataclass
class Datalink:
  route: str
  handler: Callable[[Uplink, Downlink], Downlink]

  def match(self, path: str):
    path = path[:-1] if path.endswith('/') and path != "/" else path
    return True if path == self.route else False
